from porcupine.porcupine import *
import pytest

def test_calculate_statistics():
    res_pca_pathway = pd.DataFrame({
        'pathway': ['pathway1'],
        'pathway_size': [5],
        'pc1': [3.5]
    })

    res_pca_rndm_set = pd.DataFrame({
        'pathway_size': [5, 5, 5, 5, 5],
        'pc1': [1.0, 2.0, 3.0, 4.0, 5.0]
    })
    
    result = calculate_statistics(res_pca_pathway.iloc[0], res_pca_rndm_set)

    assert result['pathway'][0] == 'pathway1'
    assert result['pathway_size'][0] == 5
    assert result['pc1'][0] == 3.5
    assert result['pval'][0] == pytest.approx(0.2592593, abs=0.00001) # R result
    assert result['es'][0] == pytest.approx(0.3162278, abs=0.00001) # R result 


def test_porcupine():
    res_pca_pathways = pd.DataFrame({
        'pathway': ['pathway1', 'pathway2'],
        'pathway_size': [5, 5],
        'pc1': [2.5, 3.0]
    })

    res_pca_rndm = pd.DataFrame({
        'pathway_size': [5, 5, 5, 5, 5, 5, 5, 5, 5, 5],
        'pc1': [1.2, 2.1, 3.1, 2.8, 1.9, 2.2, 3.0, 2.5, 1.7, 2.3]
    })
    
    result = porcupine(res_pca_pathways, res_pca_rndm)
    
    assert result['pathway'].iloc[0] == 'pathway1'
    assert result['pathway'].iloc[1] == 'pathway2'
    assert result['pathway_size'].iloc[0] == 5
    assert result['pathway_size'].iloc[1] == 5
    assert result['pc1'].iloc[0] == 2.5
    assert result['pc1'].iloc[1] == 3.0
    assert result['pval'].iloc[0] == pytest.approx(0.13651215, abs=0.00001) # R result 
    assert result['pval'].iloc[1] == pytest.approx(0.00204256 , abs=0.00001) # R result
    assert result['es'].iloc[0] == pytest.approx(0.369182, abs=0.00001) # R result
    assert result['es'].iloc[1] == pytest.approx(1.208232, abs=0.00001) # R result




import numpy as np
from .pca_pathway import *

def create_gene_set(universe, psize, n_perm=1000, seed = None, rng = None):
    """This function creates a random gene set from a universe of genes 

    Args:
        universe (set): A set of all possible genes
        psize (int): Number of genes in a gene set 
        n_perm (int): Number of permutations to create a random gene set. Defaults to 1000.
    Returns:
        pd.DataFrame : genrated pathways and theur corresponding genes.
    """    
    if rng is None:
        rng = np.random.default_rng(seed) 

    # Create a DataFrame with two columns: 'pathway' and 'genes'
    # 'pathway' contains pathway numbers ranging from 1 to n_perm
    # 'genes' contains lists of genes, each list being a random selection of psize genes from the universe
    gene_set = pd.DataFrame({
        'pathway': range(1, n_perm + 1),
        'genes': [rng.choice(list(universe), size=psize, replace=False).tolist() for _ in range(n_perm)]})
    return gene_set


def pca_random(reg_net, edges, res_pca_pathways, pathways_list, n_perm=1000, seed = None, rng = None, scale_data = True, center_data = True):
    """This function creates random gene sets and runs PCA analysis on them 

    Args:
        reg_net (pd.DataFrame) : Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): containing information on "reg" and "tar"
        res_pca_pathways (pd.DataFrame): Output result of pca_pathway function 
        pathways_list (list): A list of pathways 
        n_perm (int): Number of permutations to create a random gene set. Defaults to 1000. Defaults to 1000.
        scale_data (bool): whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool): whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        DataFrame : PCA results from random gene sets 
    """   
    # Extract the unique sizes of pathways from the PCA results 
    pathways_size = res_pca_pathways['pathway_size'].unique() 
    # Create a set of all unique genes in the provided pathways list
    universe = set(gene for genes in pathways_list['genes'] for gene in genes)
    # Initialize a list to store PCA results from random gene sets
    res_pca_random = []
    # Loop over each unique pathway size
    for psize in pathways_size:
        # Generate random gene sets of the current pathway size
        random_genes = create_gene_set(universe, psize, n_perm=n_perm, seed = seed, rng = rng)
        # Run PCA analysis on the generated random gene sets
        res_pca = pca_pathway(random_genes, reg_net, edges,scale_data=scale_data, center_data=center_data)
        # Append the PCA results to the list if it's not empty
        res_pca = res_pca[res_pca['pc1'].notna()]            
        res_pca_random.append(res_pca)

    # Concatenate all PCA results into a single DataFrame
    res_pca_random_all = pd.concat(res_pca_random, ignore_index=True)
    return res_pca_random_all



    




import os
import pandas as pd
from porcupine.run_pca import * 
from porcupine.pca_pathway import process_pathway
import pytest 

data_path = os.path.join(os.path.dirname(__file__), 'random_dataset_run_pca.csv')
data_df = pd.read_csv(data_path)

def test_run_pca():

    result = run_pca(data_df)
    
    assert result['pc1'].iloc[0] == pytest.approx(40.076, abs=0.001) #R result 
    assert result['n_edges'].iloc[0] == 10 #R result


def test_process_pathway():
    net_data = data_df
    pathway_data = pd.Series({'pathway':'PATHWAY1', 'genes' : ['gene1','gene2','gene3','gene4','gene5']})
    edges_data = {
    'reg': ['geneA', 'geneB', 'geneC', 'geneD', 'geneE'], 
    'tar': ['gene1', 'gene2', 'gene3', 'gene4', 'gene6']}
    edges_df = pd.DataFrame(edges_data)
    
    res = process_pathway(pathway_data,net_data, edges_df)

    assert res[1] == pytest.approx(48.73503, abs=0.001) #R result
    assert res[2] == 4 #R result
    assert res[3] == 5 #R result




import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

def get_ind_scores(data, scale_data=True, center_data=True):
    """This function extracts patient heterogeneity scores on the first two principal components.

    Args:
        data (pd.DataFrame): numeric matrix with samples in columns and features in rows
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): Whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        pd.DataFrame: patient heterogeneity scores for PC1 and PC2
    """
    #Transpose data
    data_t =np.transpose(data)

    # center data
    if center_data:
        mean = np.mean(data_t,axis=0)
        data_t = data_t-mean

    # scale data
    if scale_data:
        std_dev = np.std(data_t, axis=0, ddof=1)
        # avoid division by zero
        std_dev[std_dev == 0] = 1
        data_t = data_t / std_dev
        

    pca = PCA(n_components=2,svd_solver='full') # Initialize PCA object
    # Fit PCA to data
    pca.fit(data_t)
    individual_scores = pca.transform(data_t)  

    # Return a dataframe with the first two principal components
    return pd.DataFrame(individual_scores, columns=['Dim.1', 'Dim.2'], index = data.columns)


def get_pathway_ind_scores(pathways_list, reg_net, edges, scale_data=True, center_data=True):
    """This function extracts patient heterogeneity scores on the first two principal components for a list of pathways.

    Args:
        pathways_list (pd.DataFrame): list of pathways
        reg_net (pd.DataFrame): Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): containing information on "reg" and "tar"
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): Whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        pd.DataFrame: patient heterogeneity scores for each pathway
    """
    # Initialize an empty list to store results for each pathway
    res = []

    if isinstance(pathways_list, pd.DataFrame):
        pathways_iter = pathways_list.iterrows() 
    elif isinstance(pathways_list, tuple):
        pathways_iter = enumerate([pd.Series({'pathway': pathways_list[0], 'genes': pathways_list[1]})])  
    else :
        pathways_iter = enumerate([pathways_list])

    # Iterate over each pathway in the pathways list
    for _, pathway in pathways_iter:
        # Identify the indices of edges where the target genes are in the pathway's gene list
        all_genes = pathway['genes']
        idx = edges[edges['tar'].isin(all_genes)].index
        # Extract the subset of the regulatory network corresponding to these indices
        subnet = reg_net.loc[idx]
        # Perform PCA analysis on the subset
        individual_scores = get_ind_scores(subnet, scale_data=scale_data, center_data=center_data)
        # Store the results in a DataFrame with the patient as index, a pathway column and scores of the first two principal components
        res_pca = pd.DataFrame({'pathway':pathway.iloc[0], 'Dim.1':individual_scores['Dim.1'], 'Dim.2':individual_scores['Dim.2']})
        res_pca.index = individual_scores.index
        res.append(res_pca)

    res_all = pd.concat(res)   

    return res_all

def get_features_scores(data, pathway_name, reg, tar, scale_data = True, center_data = True):
    """This function extracts feature scores on the first two principal components

    Args:
        data (pd.DataFrame): Numeric matrix with samples in columns, and features in rows
        pathway_name (str): Name of the pathway
        reg (pd.Series): regulatory elements associated with the target features
        tar (pd.Series): target features
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional):  Whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        pd.DataFrame:  features scores on the PC1 and PC2
    """
    # Transpose data
    data_t = np.transpose(data)

    # Scale and/or center the data if specified
    if scale_data or center_data:
        # Initialize the StandardScaler object with mean centering and scaling
        scaler = StandardScaler(with_mean=center_data, with_std=scale_data)
        # Scale and center the data using the scaler object
        data_t = scaler.fit_transform(data_t)
    
    # Perform PCA
    pca = PCA(n_components=2)
    pca.fit(data_t)
    
    # Extract principal components (rotation matrix)
    edges_scores = pca.components_.T[:, :2]
    
    # Convert to DataFrame and add pathway, reg, and tar columns
    edges_scores = pd.DataFrame(edges_scores, columns=['pc1', 'pc2'])
    edges_scores['pathway'] = pathway_name
    edges_scores['reg'] = reg.reset_index(drop=True)
    edges_scores['tar'] = tar.reset_index(drop=True)

    return edges_scores

def get_pathway_features_scores(pathways_list, reg_net, edges, scale_data= True, center_data = True):
    """This functions extracts feature scores on the first two principal components for a list of pathways

    Args:
        pathways_list (pd.DataFrame): a list of pathways with associated genes
        reg_net (pd.DataFrame): Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): containing information on "reg" and "tar"
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): Whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        list: edges scores on PC1 and PC2 for each pathway
    """
    res = []
    for _, pathway in pathways_list.iterrows():
        pathway_name = pathway.iloc[0]
        all_genes = pathway['genes']
        idx = edges[edges['tar'].isin(all_genes)].index
        reg = edges.loc[edges['tar'].isin(all_genes), 'reg']
        tar = edges.loc[edges['tar'].isin(all_genes), 'tar']   
        # Extract the subset of the regulatory network corresponding to these indices
        subnet = reg_net.loc[idx]
        edges_scores = get_features_scores(subnet, pathway_name, reg, tar, scale_data = scale_data, center_data = center_data)
        res.append(edges_scores)

    return res 


import sys

sys.path.append('./')

from porcupine.init import *


edges_data = {
    'reg': ['geneA', 'geneB', 'geneC', 'geneD', 'geneE'],
    'tar': ['gene1', 'gene2', 'gene3', 'gene4', 'gene5']
}
edges = pd.DataFrame(edges_data)

pathways = [
    ('pathway1', ['gene1', 'gene2', 'gene3', 'gene4','gene5']),
    ('pathway2', ['gene2', 'gene3', 'gene4','gene9']),
    ('pathway3', ['gene4', 'gene5']),
    ('pathway4', ['gene10', 'gene11'])
]

def test_filter_pathways():
    pathways_filt = filter_pathways(pathways, edges)
    assert len(pathways_filt) == 3  

    

def test_filter_pathways_size():
    pathways_filt = filter_pathways(pathways, edges)
    pathways_filt_size = filter_pathways_size(pathways_filt,minSize=3,maxSize=4)
    assert len(pathways_filt_size) == 1


from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import pandas as pd
import numpy as np

def run_pca(data, scale_data=True, center_data=True):
    """This function perform PCA analysis on the data

    Args:
        data (pd.DataFrame): with samples in columns, and features in rows
        scale_data (bool): whether to scale the data (TRUE) or not (FALSE). Defaults to True
        center_data (bool):  whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        pd.DataFrame: _description_
    """    
    # Transpose the data matrix
    data_t = np.transpose(data)

    # Scale and/or center the data if specified
    if scale_data or center_data:
        # Initialize the StandardScaler object with mean centering and scaling
        scaler = StandardScaler(with_mean=center_data, with_std=scale_data)
        # Scale and center the data using the scaler object
        data = scaler.fit_transform(data_t)
    
    pca = PCA() # Initialize PCA object
    pca.fit(data) # Fit PCA to data
    
    # Calculate the percentage of variance explained by the first principal component
    pc1 = pca.explained_variance_ratio_[0] * 100
    
    result = pd.DataFrame({"pc1": pc1, "n_edges": data.shape[1]}, index=[0])
    return result





import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from yellowbrick.cluster import KElbowVisualizer
from netzoo_porcupine.pca_scores import *
import seaborn as sns
import pandas as pd
from sklearn.preprocessing import StandardScaler

def select_number_clusters(pathways_list, reg_net, edges, scale_data=True, center_data=True, kmax=8, random_state=42):
    """Visualize and select the optimal number of clusters based on PCA scores.

    Args:
        pathways_list (pd.DataFrame): List of pathways
        reg_net (pd.DataFrame): Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): DataFrame containing information on "reg" and "tar"
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): Whether to center the data (TRUE) or not (FALSE). Defaults to True.
        kmax (int, optional): Maximum number of clusters to consider. Defaults to 8.
        random_state (int, optional): Seed for random number generator for reproducibility. Defaults to 42.

    Returns:
        None: The function will plot the optimal number of clusters using the silhouette method.
    """
    # Get the PCA scores for the pathways
    pca_scores = get_pathway_ind_scores(pathways_list, reg_net, edges, scale_data=scale_data, center_data=center_data)
    
    # Extract the PCA scores for clustering
    results = pca_scores[['Dim.1', 'Dim.2']]
    
    # Initialize the KMeans model
    model = KMeans(random_state=random_state)
    
    # Initialize the KElbowVisualizer with the silhouette metric
    visualizer = KElbowVisualizer(model, k=(2, kmax + 1), metric='silhouette', timings=False)
    
    # Fit the visualizer with the data
    visualizer.fit(results)
    
    # Show the plot
    visualizer.show()

    #Save the plot
    plt.savefig('number_clusters.png', bbox_inches='tight')

    plt.close()

def visualize_clusters(pathway_of_interest, reg_net, edges, number_of_clusters, scale_data=True, center_data=True, random_state=42):
    """Visualization of clustering of patients into specified number of clusters

    Args:
        pathway_of_interest (tuple or pd.Series): List with genes in a pathway of interest
        reg_net (pd.DataFrame): Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): DataFrame containing information on "reg" and "tar"
        number_of_clusters (int): specifying the number of clusters
        scale_data (bool, optional): Whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): Whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        plot : The function will plot the clustering of patients 
    """
    # Get the PCA scores for the pathway of interest
    results = get_pathway_ind_scores(pathway_of_interest, reg_net, edges, scale_data, center_data)
    
    # Assure that results is a DataFrame
    data = results[['Dim.1', 'Dim.2']]
    
    if scale_data or center_data:
        scaler = StandardScaler(with_mean=center_data, with_std=scale_data)
        data = scaler.fit_transform(data)
    
    # Perform k-means clustering
    kmeans = KMeans(n_clusters=number_of_clusters, n_init=25, random_state=random_state)
    clusters = kmeans.fit_predict(data)
    
    # Create a DataFrame with the cluster assignments
    data_with_clusters = pd.DataFrame(data, columns=[f'Feature{i+1}' for i in range(data.shape[1])])
    data_with_clusters['Cluster'] = clusters
    
    # Plotting
    plt.figure(figsize=(6, 6))
    sns.scatterplot(x='Feature1', y='Feature2', hue='Cluster', data=data_with_clusters, palette='viridis')
    plt.title('Clusters plot')
    plt.xlabel('Dim.1')
    plt.ylabel('Dim.2')
    plt.legend(title='Cluster')
    plt.grid(True)

    #Show the plot
    plt.show()

    #Save the plot
    plt.savefig('clusters.png', bbox_inches='tight')

    plt.close()




import pyreadr
import os
import pandas as pd

def load_net_file(net_file_path=None):
    '''
    Loads network data from file specified by the user   

    Parameters:
    - net_file_path (str): path to the input networks data file 

    Returns:
        - net (DataFrame): The network data
    '''
    
    if net_file_path is None:
        net_file_path = input("Network file path => ")

    extension = os.path.splitext(net_file_path)[-1].lower()

    if extension == '.rdata':
        data = pyreadr.read_r(net_file_path)
        net = data['net'] # Network data is stored in net
    elif extension == '.tsv':
        data = pd.read_csv(net_file_path, sep='\t')
        net = data['net']
    elif extension == '.pkl':
        data = pd.read_pickle(net_file_path)
        net = data['net']
        
    return net

def load_edges_file(edges_file_path=None):
    '''
    Loads edges information from file specified by the user   

    Parameters:
    - edges_file_path (str): path to the input networks data file 

    Returns:
    - edges (DataFrame): The edges data
    '''

    if edges_file_path is None:
        edges_file_path = input("Edges file path => ")

    extension = os.path.splitext(edges_file_path)[-1].lower()

    if extension == '.rdata':
        data = pyreadr.read_r(edges_file_path)
        edges = data['edges'] # Edge data is stored in edges
    elif extension == '.tsv':
        data = pd.read_csv(edges_file_path, sep='\t')
        edges = data['edges']
    elif extension == '.pkl':
        data = pd.read_pickle(edges_file_path)
        edges = data['edges']

    return edges

def load_gmt(pathways_file_path=None):

    '''
    Load a .gmt file containing information about biological pathways

    Parameters:
    gmt_file (str): Path to the input .gmt file. Gmt file is downloaded
        from http://www.gsea-msigdb.org/gsea/msigdb/collections.jsp

    Returns:
    A list of pathways from the .gmt file
    '''
     
    if pathways_file_path is None:
        pathways_file_path = input("Pathways file path => ")

    pathways_list = []

    # Reading the .gmt file and extracting pathway and gene information
    with open(pathways_file_path, 'r') as file:
        for line in file:
            fields = line.split('\t') # Fields are tab-separated
            pathway = fields[0]  # Pathway name is the first field
            genes = [f.strip() for f in fields[2:]] # Extracting genes from fields, removing leading and trailing whitespace from each gene
            pathways_list.append((pathway, genes)) # Adding pathway name and gene list to pathways_list
    return pathways_list


def filter_pathways(pathways_list,edges):
    """
    Filter a list of pathways to include only genes present in networks

    Parameters:
    pathways_list : list of pathways
    edges : dataframe containing information on "reg" and "tar"

    Returns:
    A list of filtered pathways
    """
    
    # Extracting unique genes from edge data
    unique_edges = edges['tar'].unique()
    pathways_filt = []
    # Iterate over each pathway in the list
    for pathway, genes in pathways_list:
        filtered_genes = [gene for gene in genes if gene in unique_edges] # Filter genes to keep only those present in the unique genes from edges
        if filtered_genes:
            pathways_filt.append((pathway, filtered_genes)) # Append the pathway with filtered genes to the pathways_filt list if genes remain after filtering 

    # Creating a dataframe from patwhways_filt with columns 'pathway' and 'genes'
    pathways_filt = pd.DataFrame(pathways_filt, columns=['pathway', 'genes'])

    return pathways_filt



def filter_pathways_size(pathways_filt,minSize=5,maxSize=150):
    '''
    Filter a list of pathways based on specified minimum and maximum size for number of genes in a pathway

    Parameters:
    pathways_list : list of pathways
    minSize (int): minimum size for number of genes in a pathway (default = 5)
    maxSize (int): maximum size for number of genes in a pathway (default = 150)

    Returns:
    A list of filtered pathways
    '''
    # Filtering pathways based on the length of the gene list
    pathways_to_use = pathways_filt[
    (pathways_filt['genes'].apply(len) >= minSize) & # Filtering pathways with specified minimum size
    (pathways_filt['genes'].apply(len) <= maxSize) # Filtering pathways with specified maximum size
    ]

    return pathways_to_use

def pathways_sorting(pathways_list):
    sorted_pathways = pathways_list.sort_values(by='pathway')

    return sorted_pathways


import pandas as pd
from porcupine.pca_scores import * 
import pytest 
import os 

data_path = os.path.join(os.path.dirname(__file__), 'random_dataset_run_pca.csv')
data_df = pd.read_csv(data_path)

def test_get_features_scores():
    reg = ['reg1','reg2','reg3','reg4','reg5']
    reg = pd.Series(reg)

    tar = ['tar1','tar2','tar3','tar4','tar5']
    tar = pd.Series(tar)
    result = get_features_scores(data_df,'pathway_test', reg, tar)
    R_result = pd.DataFrame({
        'pc1': [-0.32376649, -0.24143134,-0.45551180, 0.26443944, -0.46082036, 0.06025483, -0.38935195,-0.17027459, 0.40090910, -0.04652329],
        'pc2': [0.32023960, -0.47852493, -0.20569483, -0.47162908, 0.10547065, -0.38004211, 0.14675621, -0.43796926, 0.18524924, 0.02209828]
    })

    assert np.abs(result['pc1'].iloc[0]) == pytest.approx(np.abs(R_result['pc1'].iloc[0]), abs = 0.0001)
    assert np.abs(result['pc2'].iloc[0]) == pytest.approx(np.abs(R_result['pc2'].iloc[0]), abs = 0.0001)
    assert np.abs(result['pc1'].iloc[1]) == pytest.approx(np.abs(R_result['pc1'].iloc[1]), abs = 0.0001)
    assert np.abs(result['pc2'].iloc[1]) == pytest.approx(np.abs(R_result['pc2'].iloc[1]), abs = 0.0001)


import pandas as pd
import numpy as np
from scipy.stats import ttest_1samp

def calculate_statistics(res_pca_pathway, res_pca_rndm_set):
    """This function compares the PCA score for a pathway to a set of PCA scores of random gene sets of the same size as pathway. It calculates p-value and effect size.

    Args:
        res_pca_pathway (pd.DataFrame): Result taable of pca_pathway function for one pathway
        res_pca_rndm_set (pd.DataFrame): Result table of pca_random function for one pathway

    Returns:
        pd.DataFrame: a table of statistics (p-value and effect size) for each pathway.
    """

    pc1_rndm = res_pca_rndm_set['pc1'].values
    pathway = res_pca_pathway["pathway"]
    path_size = res_pca_pathway["pathway_size"]
    pc1_pathway = res_pca_pathway["pc1"]
    
    # Calculate the p-value
    _, pvalue = ttest_1samp(pc1_rndm, popmean=pc1_pathway, alternative='less')

    # Calculate effect size with Cohen's D 
    mean_diff = abs(np.mean(pc1_rndm) - pc1_pathway)
    pooled_std = np.std(pc1_rndm, ddof=1)    
    effect_size = mean_diff / pooled_std


    
    # Store results in a DataFrame 
    stat_res = pd.DataFrame({
        "pathway": [pathway],
        "pathway_size": [path_size],
        "pc1": [pc1_pathway],
        "pval": [pvalue],
        "es": [effect_size]
    })
    
    return stat_res


def porcupine(res_pca_pathways, res_pca_rndm):
    """This function compares results of PCA analysis for each pathway versus a set of random genes sets. It calculates p-value and effect size for each pathway.

    Args:
        res_pca_pathways (pd.DataFrame): Output result table of pca_pathway function
        res_pca_rndm (pd.DataFrame): Output result table of pca_random function 

    Returns:
        pd.DataFrame: a table of statistics (pvalue and effect size) for each pathway
    """
    # Extract the unique sizes of pathways from the PCA results 
    pathways_size = res_pca_pathways['pathway_size'].unique()

    # Create an empty DataFrame
    res_all = pd.DataFrame()

    # Loop over each unique pathway size
    for path_size in pathways_size:
        print("Pathways with size", path_size)
        # Create sets with pathways of the same pathway size
        res_pathway_set = res_pca_pathways[res_pca_pathways['pathway_size'] == path_size]
        res_rndm_set = res_pca_rndm[res_pca_rndm['pathway_size'] == path_size]
        # Apply the calculate_statistics function on each pathway of the created set 
        res = res_pathway_set.apply(lambda x: calculate_statistics(x, res_rndm_set), axis=1)
        # Concatenate the DataFrmae with the obtained results
        res_all = pd.concat([res_all, pd.concat(res.tolist(), ignore_index=True)], ignore_index=True)    
        

    return res_all



    
    


    




import pandas as pd
import numpy as np
from .run_pca import *

def pca_pathway(pathways_list, reg_net, edges, scale_data=True, center_data=True):
    """
    This function runs PCA analysis for a list of pathways

    Args:
        pathways_list (pd.DataFrame): list of pathways 
        reg_net (pd.DataFrame) : Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): containing information on "reg" and "tar"
        scale_data (bool, optional): whether to scale the data (TRUE) or not (FALSE). Defaults to True
        center_data (bool, optional):  whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
        Dataframe with pca results for pathways in a pathway file
    """    
     # Initialize an empty list to store results for each pathway
    results = []
    # Iterate over each pathway in the pathways list
    for _, pathway in pathways_list.iterrows():
         # Process the pathway using PCA
        result = process_pathway(pathway, reg_net, edges, scale_data=scale_data, center_data=center_data)
        # Append the result to the results list
        results.append(result)
    # Convert the results list to a DataFrame
    res_pca = pd.DataFrame(results, columns=['pathway', 'pc1', 'n_edges', 'pathway_size'])

    return res_pca

def process_pathway(pathway, reg_net, edges, scale_data=True, center_data=True):
    """This function runs PCA analysis for one pathway 

    Args:
        pathway : a pathway with a list of genes associated 
        reg_net (pd.DataFrame) : Numeric matrix with samples in columns, features in rows
        edges (pd.DataFrame): containing information on "reg" and "tar"
        scale_data (bool, optional):whether to scale the data (TRUE) or not (FALSE). Defaults to True.
        center_data (bool, optional): whether to center the data (TRUE) or not (FALSE). Defaults to True.

    Returns:
       list: A list containing:
              - pathway : The input pathway 
              - pc1 (float): The first principal component score from the PCA analysis.
              - n_edges (int): The number of edges used in the PCA analysis.
              - n_genes (int): The number of genes in the pathway.
    """    
    # Identify the indices of edges where the target genes are in the pathway's gene list
    all_genes = (pathway['genes'])
    idx = edges[edges['tar'].isin(all_genes)].index
     # Extract the subset of the regulatory network corresponding to these indices
    subnet = reg_net.loc[idx]
    pc1 = None
    n_edges = None
    # Perform PCA analysis on the subset
    if not subnet.empty:
        pca_result = run_pca(subnet, scale_data=scale_data, center_data=center_data)
        pc1 = pca_result['pc1'].iloc[0]
        n_edges = pca_result['n_edges'].iloc[0]
 
    return (pathway.iloc[0], pc1, n_edges, len(pathway['genes']))